from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import argparse
import sys
import numpy as np 
import tensorflow as tf 
from tensorflow.contrib.layers import fully_connected
import random
from tensorflow.contrib.layers import batch_norm
from scipy import stats
from scale_noise import augment_noise_scale



# expr: the n*m (cell*gene) expression matrix
# labels: one-hot encoded group labels (n*k/cell*group)
def mean_std_est(expr,labels):
    n_cl = labels.shape[1]  #n_cl is the number of classes
    n_gene = expr.shape[1]
    U = np.mean(expr,axis=0) # the mean expression level of each gene (mean of each column)
# estimate the mean and standard deviation within each class
    Mu = np.zeros([n_gene,n_cl]) 
    Sigma = np.zeros([n_gene,n_cl])
    for i in range(n_cl):
        Mu[:,i] =np.mean(expr[(labels[:,i]==1),],axis=0)
        Sigma[:,i] = np.sum((expr[(labels[:,i]==1),].T-Mu[:,i].reshape([n_gene,1]))*(expr[(labels[:,i]==1),].T-Mu[:,i].reshape([n_gene,1])),axis=1)

    std = np.sqrt(np.sum(Sigma,axis=1)/(expr.shape[0]-n_cl))

    ### fit a line to the data: x-U, y-std
    slope,intercept,r_value,p_value,slope_std_error=stats.linregress(U,std)
    std_h = intercept+slope*U # the predicted standard deviation

    return U,std_h

####


def augment(expr,labels,Fold=20,alpha=0.2):
    n_cell = expr.shape[0]
    # expr1 is the augmented expression matrix:
    expr1 = np.tile(expr,(Fold,1))
    U,std_h = mean_std_est(expr,labels)
    Var = np.random.multivariate_normal(mean=np.zeros(expr.shape[1]),cov=alpha*np.diag(np.square(std_h)),size=Fold*n_cell)
    expr1 = expr1+Var
    expr1 = np.vstack((expr,expr1))
    labels1=np.tile(labels,(Fold+1,1))
    #orders = np.random.choice(a=expr1.shape[0],size=expr1.shape[0],replace=False)
    #expr1=expr1[orders,]
    #labels1=labels1[orders,]
    
    return expr1,labels1
###################################################################################

def aug_test(expr,U_min,U_max,Fold=1000):
    n_cell,n_genes = expr.shape  
    U1 =-2
    U2 = 2
    ## (u1,u2) is the range to generate scale factor
    SF = np.zeros(n_cell*Fold)
    SF = np.random.uniform(U1,U2,size=n_cell*Fold)   
    SF = SF.reshape((n_cell*Fold,1))
    expr1 = np.tile(expr,(Fold,1))+SF # expr1 is the scale adjusted expression matrix
    expr1 = np.vstack((expr,expr1))
    
    return expr1 
#######################################################################
# sdc() is a function which build, train and test the network.
#######################################################################
## Argurments:
# expr_train, labels_train: the training data set; 
#      expr_train: the expression data of training sets on the log scale;
#      labels_train: the class labels of training data.
# expr_test, labels_test: the test dataset;
#       expr_test: the expression data of test sets on the log scale;
#       labels_test: the class labels of test data.
# epoch: the maximal number of training epochs, namely the folds of augmentation.
# min_epoch: the minimal number of training epochs
# lr: the starting value of learning rate to train the network.
# batch_size: the batch size to train the network.
# min_stop: when the training epoch is larger than min_stop, and the change of loss is smaller than min_stop, stop the training.    
# display_step: for every display_step of epochs, print the current number of epochs, loss value, and classification accuracy on the test data.
# aug: the folds of augmentation for test data to evaluate the accuracy on the augmented test data and concordance rate.
## Values:
# res0: classification accuracy on the test data
# res1: classification accuracy on the augmented test data 
# OP: the concordance rate of the predicted labels on the augmented test data
# clist: a list that records the value of loss after each training epoch
# ct_list: a list that records the test accuracy after each trainig epoch
# pred_test: the output of the trained network on the test data; 
#           pred_test is a N_test * K matrix, where N_test is the sample size of test data, K is the number of classes
#           the predicted label of test data can be obtained by np.argmax(pred_test).

def sdc(expr_train, labels_train,
        expr_test,labels_test,
        epoch = 1000, 
        lr=0.05,
        batch_size=32,
        min_stop = 0.01,
        min_epoch=300,
        display_step=100,
        aug=500): # aug is the augmentation folds of test data

# if aug==True, then augment training data
    tf.set_random_seed(400)
    print("The sample size of training data is: ",expr_train.shape)
    print("The learning rate is: ",lr)
    print("The batch size is:",batch_size)
    cl_dim = labels_train.shape[1]
    n_cell,in_dim=expr_train.shape
# Placeholder for input data
    X = tf.placeholder(tf.float32,[None,in_dim])
    Y = tf.placeholder(tf.float32,[None,cl_dim])
    keep_prob_big = tf.placeholder(tf.float32) # for the dropout layer
    keep_prob_small = tf.placeholder(tf.float32)
# Define the structure of the network
# batch normalization
    is_training = tf.placeholder(tf.bool,shape=(),name='is_training')

    bn_params = {
        'is_training': is_training,
        'decay': 0.99,
        'scale': True,
        'updates_collections':None,
        'fused':False,
    }
# 4 hidden layers
    X_drop = tf.nn.dropout(X,keep_prob=keep_prob_big)
    h0 = fully_connected(X,256,activation_fn = tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h0 = tf.nn.dropout(h0,keep_prob=keep_prob_small)
    h1 = fully_connected(X,128,activation_fn = tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h1 = tf.nn.dropout(h1,keep_prob=keep_prob_small)
    h2 = fully_connected(h1,64,activation_fn=tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h2 = tf.nn.dropout(h2,keep_prob=keep_prob_small)
    h3 = fully_connected(h2,32,activation_fn=tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h3 = tf.nn.dropout(h3,keep_prob=keep_prob_small)
    h4 = fully_connected(h3,16,activation_fn=tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h4 = tf.nn.dropout(h4,keep_prob=keep_prob_big)
    h5 = fully_connected(h4,8,activation_fn=tf.nn.relu,normalizer_fn=batch_norm,normalizer_params=bn_params)
    h5 = tf.nn.dropout(h5,keep_prob=keep_prob_big)
    print("The dimension of 6 layers are:",h0.shape[1],h1.shape[1],h2.shape[1],h3.shape[1],h4.shape[1],h5.shape[1])
    pred = fully_connected(h1,cl_dim,activation_fn=tf.nn.softmax,normalizer_fn=batch_norm,normalizer_params=bn_params)
# Define loss as cross entropy
    cross_entropy = tf.reduce_mean(-tf.reduce_sum(Y*tf.log(pred),reduction_indices=[1]))

# Define the optimization algorithm as AdamOptimzer
    global_step=tf.Variable(0)
    lr_c = tf.placeholder(tf.float32)
    learning_rate = tf.train.exponential_decay(lr_c,global_step,decay_steps=int(n_cell/batch_size),decay_rate=0.9,staircase=True)
    train_step = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cross_entropy,global_step=global_step)

    with tf.Session() as sess:
        c0 = 1000000.0
        init = tf.global_variables_initializer()
        sess.run(init)
        total_batch = int(n_cell/batch_size)
        correct_predict = tf.equal(tf.argmax(Y,1),tf.argmax(pred,1))
        accuracy = tf.reduce_mean(tf.cast(correct_predict,tf.float32))
        Scales = np.mean(expr_test,axis=1) # scale is estimated as the sum of each cell
        expr_aug = aug_test(expr_test,U_min=min(Scales),U_max=max(Scales),Fold=aug)
        labels_aug = np.tile(labels_test,(aug+1,1))
        clist=[]
        ct_list=[]
        for e in range(epoch):
            expr1,SF_aug = augment_noise_scale(expr_train,labels_train,Fold=2,alpha=1)
            expr_train1 = np.vstack((expr_train,expr1))
            arr = np.arange(n_cell*2)
            np.random.shuffle(arr)
            for i in range(total_batch):
                batch_indx = arr[(i*batch_size):(i*batch_size+batch_size-1)]
                batch_xs = expr_train1[batch_indx%n_cell,]+SF_aug[batch_indx,]
                batch_ys = labels_train[batch_indx%n_cell,]
                
                _, c=sess.run([train_step,cross_entropy],feed_dict={X:batch_xs,Y:batch_ys,lr_c:lr,keep_prob_big:1.0,keep_prob_small:0.5,is_training:True})
                
            acc_test = sess.run(accuracy,feed_dict={X:expr_aug,Y:labels_aug,keep_prob_big:1.0,keep_prob_small:1.0,is_training:False})
            clist.append(c)
            ct_list.append(acc_test)
            if e % display_step ==0:
                print("Epoch:",'%04d'%(e+1),"cost=","{:.9f}".format(c),"test_acc=",acc_test)
            if (c)<min_stop and (e)>(min_epoch):
                print("previous cost",c0,"current cost:",c)
                break
            c0 = c
        print("Optimization Finished!\n")


    
        pred_test = sess.run(pred,feed_dict={X:expr_test,keep_prob_big:1.0,keep_prob_small:1.0,is_training:False})
        res0=(sess.run(accuracy,feed_dict={pred:pred_test,Y:labels_test}))
        res1 = sess.run(accuracy,feed_dict={X:expr_aug,Y:labels_aug,keep_prob_big:1.0,keep_prob_small:1.0,is_training:False})
        res_b = 0.0
        n_test = expr_test.shape[0]  # the size of test set
        OP = np.zeros(n_test)
        for i in range(n_test):
            expr_test_b = expr_test[i,].reshape([1,expr_test.shape[1]])  # the i-th row(sample) of test set
            expr_test_b = aug_test(expr_test_b,U_min=min(Scales),U_max=max(Scales),Fold=aug) # augment of scaling for sample i
            pred_b=sess.run(pred,feed_dict={X:expr_test_b,keep_prob_big:1.0,keep_prob_small:1.0,is_training:False})
            labels_sum = np.sum(pred_b,axis=0)
            labels_b = np.zeros([expr_test_b.shape[0],labels_test.shape[1]])
            labels_b[:,np.argmax(labels_sum)]=1
            OP[i]=sess.run(accuracy,feed_dict={pred:pred_b,Y:labels_b})

        print("The concordance rate is ",np.mean(OP))
        print("classification accuracy of original test data is ",res0)
    
    return res0,res1,OP,clist,ct_list,pred_test 
